@@ -29,8 +29,7 @@ const GetResultType = {
     build: 1,
     errors: 2,
     command: 3,
-}
-
+};
 
 // A regex to match a percentage: one or more digits, optionally followed by a period and one or more digits, and ending with a percent sign.
 let percent_regex = new RegExp(
@@ -40,16 +39,18 @@ let percent_regex = new RegExp(
         "\\d+" +
         // Optionally followed by a decimal point and one ore more digits. (Don't include these in a capture group.)
         "(?:\\.\\d+)?" +
-    // End the capture group, then require a percent sign. Search globally (for all matches).
-    ")%", "g"
+        // End the capture group, then require a percent sign.
+        ")%",
+    // Search globally (for all matches).
+    "g"
 );
 
-
 // Core client
 // ===========
 // Given an ID to use, run the CodeChat Client.
-function run_client(id, ws_address)
-{
+function run_client(id, ws_address) {
+    // Set up variables used by the functions below
+    // --------------------------------------------
     // Get commonly-used nodes in the DOM.
     let status_message = document.getElementById("status_message");
     let build_progress = document.getElementById("build_progress");
@@ -59,8 +60,6 @@ function run_client(id, ws_address)
     let build_contents = document.getElementById("build_contents");
     let errors_div = document.getElementById("errors");
 
-    // Set up "globals" for the function below.
-    //
     // True if the output/errors should be cleared on the next result.
     let clear_output = true;
     // True if the most recent reload was caused by the user; false if this was a programmatic reload.
@@ -71,7 +70,10 @@ function run_client(id, ws_address)
     let splitter_size = {
         true: 85,
         false: 100,
-    }
+    };
+
+    // Core code
+    // ---------
     splitMe.init();
     set_splitter_percent(splitter_size[errors_or_warnings]);
 
@@ -80,24 +82,30 @@ function run_client(id, ws_address)
 
     // Identify this client on connection.
     ws.onopen = () => {
-        console.log(`CodeChat Client: websocket to CodeChat Server open. Sending ID of ${id}.`);
-        ws.send(JSON.stringify(id))
+        console.log(
+            `CodeChat Client: websocket to CodeChat Server open. Sending ID of ${id}.`
+        );
+        ws.send(JSON.stringify(id));
     };
 
     // Provide logging to help track down errors.
-    ws.onerror = event => {
+    ws.onerror = (event) => {
         console.error(`CodeChat Client: websocket error ${event}.`);
     };
 
-    ws.onclose = event => {
-        console.log(`CodeChat Client: websocket closed by event ${event}. This should only happen on shutdown.`);
-    }
+    ws.onclose = (event) => {
+        console.log(
+            `CodeChat Client: websocket closed by event ${event}. This should only happen on shutdown.`
+        );
+    };
 
     // Handle messages.
-    ws.onmessage = event => {
+    ws.onmessage = (event) => {
         result = JSON.parse(event.data);
         if (result.get_result_type === GetResultType.url) {
-            console.log(`CodeChat Client: URL ${result.text} received; loading...`);
+            console.log(
+                `CodeChat Client: URL ${result.text} received; loading...`
+            );
             // Save and restore scroll location through the content update, if we can.
             let [scrollX, scrollY] = getScroll();
             // See ideas in https://stackoverflow.com/a/16822995. Works for same-domain only.
@@ -111,7 +119,7 @@ function run_client(id, ws_address)
                     build_progress.style.display = "none";
                     outputElement.classList.remove("building");
                     // Restore the scroll location if we were able to save it.
-                    if ( (scrollX !== undefined) && (scrollY !== undefined) ) {
+                    if (scrollX !== undefined && scrollY !== undefined) {
                         this.contentWindow.scrollBy(scrollX, scrollY);
                     }
                     // The programmatic reload is done -- anything else that happens now is from the user.
@@ -128,13 +136,20 @@ function run_client(id, ws_address)
 
             // Set the new src to (re)load content. At startup, the ``srcdoc`` attribute shows some welcome text. Remove it so that we can now assign the ``src`` attribute.
             outputElement.removeAttribute("srcdoc");
-            outputElement.src = window.location.protocol + '//' + window.location.host + window.location.pathname + "/" + id + "/" + result.text;
+            outputElement.src =
+                window.location.protocol +
+                "//" +
+                window.location.host +
+                window.location.pathname +
+                "/" +
+                id +
+                "/" +
+                result.text;
             // The next build output received will apply to the new build, so set the flag.
             clear_output = true;
             // See comments above -- avoid double loads and indicate that the is a programmatic reload.
             is_user_navigation = false;
             // Exit for now; the callbacks above will continue this function's operation.
-
         } else if (result.get_result_type === GetResultType.build) {
             // Omit this logging, since it's usually obvious from the lines below.
             ///console.log("CodeChat Client: received build output.");
@@ -161,11 +176,12 @@ function run_client(id, ws_address)
             let percent_matches = result.text.match(percent_regex);
             if (percent_matches !== null) {
                 // Update the progress bar with the percentage from the last match.
-                build_progress.value = parseFloat(percent_matches[percent_matches.length - 1]);
+                build_progress.value = parseFloat(
+                    percent_matches[percent_matches.length - 1]
+                );
             }
             // Scroll to the bottom, to show the content just added.
             scroll_to_bottom(build_contents);
-
         } else if (result.get_result_type === GetResultType.errors) {
             console.log("CodeChat Client: error output received.");
             if (clear_output) {
@@ -182,59 +198,95 @@ function run_client(id, ws_address)
             scroll_to_bottom(build_contents);
 
             // Update the errors/warnings and splitter position.
-            [errors_div.innerHTML, status_errors_div.innerHTML, errors_or_warnings] =
-                parse_for_errors(errors_div.innerHTML);
+            [
+                errors_div.innerHTML,
+                status_errors_div.innerHTML,
+                errors_or_warnings,
+            ] = parse_for_errors(errors_div.innerHTML);
             set_splitter_percent(splitter_size[errors_or_warnings]);
-
         } else if (result.get_result_type === GetResultType.command) {
             console.log(`CodeChat Client: command ${result.text} received.`);
             if (result.text === "shutdown") {
                 // Close this window -- see https://stackoverflow.com/a/54787080. See `close the window`_.
-                outputElement.srcdoc = "<html><body><p>The CodeChat Client was shut down. Please close this window.</p></body></html>";
+                outputElement.srcdoc =
+                    "<html><body><p>The CodeChat Client was shut down. Please close this window.</p></body></html>";
                 build_div.textContent = "";
                 errors_div.textContent = "";
                 status_message.innerHTML = "Client shut down.";
-                window.open('', '_self').close();
+                window.open("", "_self").close();
                 // Stop asking for results.
                 ws.close();
                 console.log("CodeChat Client: shutdown complete.");
             } else if (result.text.startsWith("error: unknown client ")) {
                 // _`Close the window` -- there's no point in asking for more commands. Note: there are some cases where this fails, although I've only seen this once. From the Chrome console, ``Scripts may close only the windows that were opened by them.`` In this case, leave a message asking the user to close the window.
-                outputElement.srcdoc = "<html><body><p>CodeChat Client error: lost connection to the CodeChat Server. Close this window and restart the editor plug-in.</p></body></html>";
+                outputElement.srcdoc =
+                    "<html><body><p>CodeChat Client error: lost connection to the CodeChat Server. Close this window and restart the editor plug-in.</p></body></html>";
                 build_div.textContent = "";
                 errors_div.textContent = "";
                 status_message.innerHTML = "Server disconnected";
-                window.open('', '_self').close();
+                window.open("", "_self").close();
                 // Stop asking for results.
                 ws.close();
-                console.log(`CodeChat Client: shutdown complete due to unknown client; saw ${result.text}.`);
+                console.log(
+                    `CodeChat Client: shutdown complete due to unknown client; saw ${result.text}.`
+                );
             } else {
                 console.log(`CodeChat Client: Unknown command ${result.text}.`);
             }
-
         } else {
-            console.log(`CodeChat Client: Unknown GetResultType: ${result.get_result_type}.`);
+            console.log(
+                `CodeChat Client: Unknown GetResultType: ${result.get_result_type}.`
+            );
         }
     };
 
+    // Provide utility functions for use by the core functions above
+    // -------------------------------------------------------------
+
     // Return the X and Y coordinates of the scrollbar of the output iframe if possible.
     function getScroll() {
         try {
             // This is only allowed for same-domain origins -- for example, if the user clicks on a link in the docs that goes to an external website, then the following lines will raise an exception.
-            return [outputElement.contentWindow.scrollX, outputElement.contentWindow.scrollY];
+            return [
+                outputElement.contentWindow.scrollX,
+                outputElement.contentWindow.scrollY,
+            ];
         } catch (err) {
             return [undefined, undefined];
         }
     }
 
-    // The let statement below makes this accessible globally.
-    navigate_to_error = function(file_path, line) {
-        // TODO.
-        console.log("CodeChat Client: TODO", file_path, line);
+    // Send a message to the CodeChat server.
+    function send_to_codechat_server(
+        // A string, indentifying the type of message.
+        msg,
+        // Message-specific data to send. Must be JSON-encodable.
+        data
+    ) {
+        ws.send(JSON.stringify([msg, data]));
+    }
+
+    // These functions define _`messages sent by the CodeChat Client`, which are handled by `read_websocket_handler <read_websocket_handler>`. Defining them this way makes them accessible to the iframe and globally (to functions called outside this function). See below.
+    //
+    // Save data to the filesystem. Right now, only used for pretext editing.
+    save_file = function (xml_node, file_contents) {
+        send_to_codechat_server("save_file", {
+            xml_node: xml_node,
+            file_contents: file_contents,
+        });
+    };
+
+    // The let statements below makes these accessible globally.
+    navigate_to_error = function (file_path, line) {
+        send_to_codechat_server("navigate_to_error", {
+            file_path: file_path,
+            line: line,
+        });
     };
 }
 
-let navigate_to_error;
+// Globally-accessible functions (see above). These must be declared as ``var`` (not ``let``) due to (I assume) scoping rules I don't understand.
+var navigate_to_error, save_file;
 
 // Utilities
 // =========
@@ -243,7 +295,6 @@ function get_splitter() {
     return document.getElementById("splitter");
 }
 
-
 // Set the splitter percentage
 function set_splitter_percent(percent) {
     let splitter = get_splitter();
@@ -251,13 +302,11 @@ function set_splitter_percent(percent) {
     splitMe.update(splitter);
 }
 
-
 // Scroll an element to the bottom
 function scroll_to_bottom(element) {
     element.scrollTop = element.scrollHeight - element.clientHeight;
 }
 
-
 // This regex parses the error string to determine get the number of warnings and errors.
 const error_regex = new RegExp(
     // Common docutils error messages read::
@@ -277,7 +326,7 @@ const error_regex = new RegExp(
     //  In Sphinx 1.6.1:
     //  X:\file.rst: WARNING: document isn't included in any toctree
     //
-    // The CodeChat renderer also produces `error messages <_checkModificationTime>`_ formatted in a similar way so they'll be identified by the same regex::
+    // The CodeChat renderer also produces `error messages <checkModificationTime>`_ formatted in a similar way so they'll be identified by the same regex::
     //
     //  X:\ode.py:: ERROR: CodeChat renderer - source file older than the html file X:\_build\html\ode.py.
     //
@@ -290,51 +339,53 @@ const error_regex = new RegExp(
     //
     // The filename is anything up to the colon. Windows filenames may begin with a drive letter followed by a colon -- don't capture this (the leading ``?:``).
     "^((?:\\w:)?[^:]*)" +
-    // Find the first occurrence of a pair of colons, or just a single colon. Between them there can be numbers or "None" or nothing. For example, this expression matches the string ":1589:" or string ":None:" or the string "::" or the string ":".
-    ":(\\d*|None):? " +
-
-    // Next match the error type, which can only be "WARNING", "ERROR" or "SEVERE". Before this error type the message may optionally contain one left parenthesis.
-    "\\(?(WARNING|ERROR|SEVERE)" +
-
-    // Since one error message occupies one line, a ``*`` quantifier is used along with end-of-line ``$`` to make sure only the first match is used in each line.
-    ".*$",
+        // Find the first occurrence of a pair of colons, or just a single colon. Between them there can be numbers or "None" or nothing. For example, this expression matches the string ":1589:" or string ":None:" or the string "::" or the string ":".
+        ":(\\d*|None):? " +
+        // Next match the error type, which can only be "WARNING", "ERROR" or "SEVERE". Before this error type the message may optionally contain one left parenthesis.
+        "\\(?(WARNING|ERROR|SEVERE)" +
+        // Since one error message occupies one line, a ``*`` quantifier is used along with end-of-line ``$`` to make sure only the first match is used in each line.
+        ".*$",
 
     // The message usually contains multiple lines; search each line for errors and warnings.
     "m" +
-    // The global flag must be present to replace all occurrences.
-    "g"
+        // The global flag must be present to replace all occurrences.
+        "g"
 );
 
-
 // Parse the error output for errors and warnings.
 function parse_for_errors(errors_html) {
     let errNum = 0;
     let warningNum = 0;
     // The replacement function is called with the match text then each matched group.
-    errors_html = errors_html.replace(error_regex, function (match_text, file_path, line, error_string) {
-        if ((error_string === "ERROR") || (error_string === "SEVERE")) {
-            ++errNum;
-        } else if (error_string === "WARNING") {
-            ++warningNum;
-        }
+    errors_html = errors_html.replace(
+        error_regex,
+        function (match_text, file_path, line, error_string) {
+            if (error_string === "ERROR" || error_string === "SEVERE") {
+                ++errNum;
+            } else if (error_string === "WARNING") {
+                ++warningNum;
+            }
 
-        // Unsanitize the file name.
-        let div = document.createElement("div");
-        div.innerHTML = file_path;
-        file_path = div.textContent;
-        // Make an unspecified filename empty.
-        if (file_path === "<string>") {
-            file_path = "";
-        }
+            // Unsanitize the file name.
+            let div = document.createElement("div");
+            div.innerHTML = file_path;
+            file_path = div.textContent;
+            // Make an unspecified filename empty.
+            if (file_path === "<string>") {
+                file_path = "";
+            }
 
-        // Clean up the line.
-        if ((line === "None") || (line === "")) {
-            line = -1;
-        }
+            // Clean up the line.
+            if (line === "None" || line === "") {
+                line = -1;
+            }
 
-        // Create a hyperlink to navigate to the error.
-        return `<a href='javascript:navigate_to_error(${JSON.stringify(file_path)}, ${line})' class="error_link">${match_text}</a>`;
-    });
+            // Create a hyperlink to navigate to the error.
+            return `<a href='javascript:navigate_to_error(${JSON.stringify(
+                file_path
+            )}, ${line})' class="error_link">${match_text}</a>`;
+        }
+    );
 
     // Report these results to the user.
     let span_class;
@@ -348,6 +399,6 @@ function parse_for_errors(errors_html) {
     return [
         errors_html,
         `<span class="${span_class}">Error(s): ${errNum}, warning(s): ${warningNum}</span>`,
-        errNum + warningNum > 0
+        errNum + warningNum > 0,
     ];
 }
