@@ -241,7 +241,7 @@ def start(
             print(p.stderr.read())
             print("The server exited unexpectedly.", file=sys.stderr)
             return 1
-        sleep(0.1)
+        sleep(0.001)
     # Indicate the server started to the timeout thread, so the timeout won't kill the server now that it's started.
     server_starting = False
     return 0
@@ -312,9 +312,13 @@ def build(path_to_build: List[Path]) -> int:
     for ptb in path_to_build:
         ptb = ptb.resolve()
         print(f"Building {ptb}...", file=sys.stderr)
-        was_performed, rendered_file_path, html, err_string = asyncio.run(
-            render_file(file_text(ptb), str(ptb), aprint, False)
-        )
+        (
+            was_performed,
+            rendered_project_path,
+            rendered_file_path,
+            html,
+            err_string,
+        ) = asyncio.run(render_file(file_text(ptb), str(ptb), aprint, False))
         assert was_performed
         # Print all errors produced by the render.
         print(err_string, file=sys.stderr)
