@@ -41,15 +41,17 @@ from typing import (
     Union,
 )
 import urllib.parse
+import xml.dom.minidom
 
 # Third-party imports
 # -------------------
+from lxml import etree as ElementTree
 import websockets
 
 # Local imports
 # -------------
 from .constants import LOCALHOST, WEBSOCKET_PORT
-from .renderer import render_file, is_win
+from .renderer import is_win, ProjectConfFile, render_file
 
 
 # RenderManager / render thread
@@ -368,14 +370,67 @@ class RenderManager:
 
             msg, data = json.loads(ret)
             if msg == "save_file":
-                print(f"Save to {data['xml_node']} values '{data['file_contents']}'.")
+                print(
+                    f"Save to {data['xml_node']} values:\n{data['file_contents'][:77]}..."
+                )
                 # Get the location of the project file.
                 pp = self._client_state_dict[id_]._project_path
                 if not pp:
                     print("Unable to save: no project file available.")
                     continue
-
-                # The pretext build must provide the root document and a map of {source file path: [rendered file paths]} for all files in a build. Invert this to compute the source file path given a rendered file path.
+                # Read the source path from it.
+                project_conf = ProjectConfFile(Path(pp))
+                # Find the source file which matches this mapping.
+                xml_id_to_replace = data["xml_node"]
+                for (
+                    source_file,
+                    xml_id_list,
+                ) in project_conf.load_pretext_mapping().items():
+                    if xml_id_to_replace in xml_id_list:
+                        try:
+                            # Load in this source.
+                            src_tree = ElementTree.parse(source_file)
+                            # Parse the replacement source for it.
+                            new_node = ElementTree.fromstring(data["file_contents"])
+                        except Exception as e:
+                            print(f"Unable to load file or parse new source: {e}")
+                            break
+                        # Find the node in this source file and replace it. Find only looks at children, so manually check the root element.
+                        xml_id_attrib = "{http://www.w3.org/XML/1998/namespace}id"
+                        xml_node_to_replace = (
+                            src_tree.getroot()
+                            if src_tree.getroot().get(xml_id_attrib)
+                            == xml_id_to_replace
+                            else src_tree.find(
+                                f"//*[@{xml_id_attrib} = '{xml_id_to_replace}']"
+                            )
+                        )
+                        if xml_node_to_replace is None:
+                            print(
+                                f"Unable to save: can't find node {xml_node_to_replace} in {source_file}."
+                            )
+                            break
+                        # The correct method to replace this node is either ``_setroot`` (this is is the root element) or ``replace`` (otherwise).
+                        parent = xml_node_to_replace.getparent()
+                        if parent is None:
+                            src_tree._setroot(new_node)
+                        else:
+                            parent.replace(xml_node_to_replace, new_node)
+                        # Save the updated file. I can't find a way using lxml (including looking at the `docinfo <https://lxml.de/api/lxml.etree.DocInfo-class.html>`_ class) to determine if the loaded document contained an XML declaration (the ``<?xml version="1.0" encoding="UTF-8"?>`` header) or not. The following comes from `SO <https://stackoverflow.com/a/54942605/16038919>`__.
+                        has_xml_declaration = bool(
+                            xml.dom.minidom.parse(source_file).version
+                        )
+                        src_tree.write(
+                            source_file,
+                            encoding="utf-8",
+                            xml_declaration=has_xml_declaration,
+                        )
+                        print(f"Saved to {source_file}.")
+                        break
+                else:
+                    print(
+                        f"Unable to write: can't find source file containing {xml_id_to_replace}."
+                    )
             elif msg == "navigate_to_error":
                 # TODO
                 print(
