@@ -33,8 +33,10 @@
 import asyncio
 import codecs
 from contextlib import contextmanager
+from enum import Enum, auto
 import fnmatch
 import io
+import json
 from pathlib import Path
 import shlex
 import sys
@@ -230,86 +232,214 @@ async def _render_external_file(
 
 # Project
 # -------
+# Define the possible types of projects.
+class ProjectTypeEnum(Enum):
+    general = auto()
+    PreTeXt = auto()
+    Doxygen = auto()
+
+
+# This class reads and interprets a project configuration file.
+class ProjectConfFile:
+    # Read and process a CodeChat project configuration file.
+    def __init__(self, project_path: Path):
+        try:
+            with open(project_path, encoding="utf-8") as f:
+                data = f.read()
+        except Exception as e:
+            raise RuntimeError(f"{project_path}:: ERROR: Unable to open. {e}")
+
+        schema = strictyaml.Map(
+            {
+                strictyaml.Optional("source_path", default="."): strictyaml.Str(),
+                "output_path": strictyaml.Str(),
+                "args": strictyaml.Str() | strictyaml.Seq(strictyaml.Str()),
+                strictyaml.Optional("html_ext", default=".html"): strictyaml.Str(),
+                strictyaml.Optional("project_type", default="general"): strictyaml.Enum(
+                    [x.name for x in list(ProjectTypeEnum)]
+                ),
+            }
+        )
+        try:
+            data_dict = strictyaml.load(data, schema).data
+        except strictyaml.YAMLError as e:
+            raise RuntimeError(f"{project_path}:: ERROR: Unable to parse. {e}")
+
+        # Save items that don't need processing.
+        self.project_type = ProjectTypeEnum[data_dict["project_type"]]
+        self.html_ext = data_dict["html_ext"]
+
+        # Make paths absolute.
+        project_path = project_path.parent
+
+        def abs_path(path: Union[str, Path]) -> Path:
+            path_ = Path(path)
+            if not path_.is_absolute():
+                path_ = project_path / path_
+            return path_
+
+        self.source_path = abs_path(data_dict["source_path"])
+        self.output_path = abs_path(data_dict["output_path"])
+
+        # Perform replacement on the args.
+        def args_format(arg):
+            return arg.format(
+                project_path=project_path,
+                source_path=self.source_path,
+                output_path=self.output_path,
+            )
+
+        args = data_dict["args"]
+        self.args = (
+            args_format(args)
+            if isinstance(args, str)
+            else [args_format(arg) for arg in args]
+        )
+
+    # _`checkModificationTime`: Verify that ``source_file``` is older than the HTML file produced by this source, reporting errors if this isn't true.
+    def checkModificationTime(
+        self, source_file: Path
+    ) -> Tuple[
+        # A path to the proposed or found HTML file.
+        str,
+        # A list of error messages.
+        List[str],
+    ]:
+
+        # Determine a first guess at the location of the rendered HTML.
+        prefix = (
+            self.output_path
+            if self.project_type != ProjectTypeEnum.Doxygen
+            else Path(".")
+        )
+        error_arr = []
+        try:
+            base_html_file = prefix / source_file.relative_to(self.source_path)
+        except ValueError as e:
+            # Give some arbitrary value to the output path, since it can't be determined.
+            base_html_file = prefix
+            error_arr.append(
+                "{}:: ERROR: unable to compute path relative to {}. {}".format(
+                    source_file, self.source_path, e
+                )
+            )
+
+        # For PreTeXt, use the mapping of source files to XML IDs == output file name.
+        if self.project_type == ProjectTypeEnum.PreTeXt:
+            try:
+                mapping = self.load_pretext_mapping()
+            except Exception:
+                pass
+            else:
+                # Before looking up the file, ``resolve()`` it to get the canonical representation (fix case on Windows).
+                if xml_id_list := mapping.get(str(source_file.resolve())):
+                    # Pick the first mapping, since there may be several.
+                    base_html_file = self.output_path / xml_id_list[0]
+        # For Doxygen, rename certain characters in the file name. See `util.cpp::escapeCharsInString <https://github.com/doxygen/doxygen/blob/master/src/util.cpp#L3443>`_.
+        elif self.project_type == ProjectTypeEnum.Doxygen:
+            doxygen_renamed_path = base_html_file.as_posix()
+            for old, new in (
+                [":", "_1"],
+                ["/", "_2"],
+                ["<", "_3"],
+                [">", "_4"],
+                ["*", "_5"],
+                ["&", "_6"],
+                ["|", "_7"],
+                [".", "_8"],
+                ["!", "_9"],
+                [",", "_00"],
+                [" ", "_01"],
+                ["{", "_02"],
+                ["}", "_03"],
+                ["?", "_04"],
+                ["^", "_05"],
+                ["%", "_06"],
+                ["(", "_07"],
+                [")", "_08"],
+                ["+", "_09"],
+                ["=", "_0a"],
+                ["$", "_0b"],
+                ["\\", "_0c"],
+                ["@", "_0d"],
+                ["]", "_0e"],
+                ["[", "_0f"],
+                ["#", "_0g"],
+            ):
+                doxygen_renamed_path = doxygen_renamed_path.replace(old, new)
+            base_html_file = self.output_path / doxygen_renamed_path
+
+        # Look for the resulting HTML using this guess.
+        possible_html_file = base_html_file.with_suffix(self.html_ext)
+        if possible_html_file.exists():
+            html_file = possible_html_file
+        else:
+            possible_html_file = Path(str(base_html_file) + self.html_ext)
+            if possible_html_file.exists():
+                html_file = possible_html_file
+            else:
+                return (
+                    str(base_html_file),
+                    error_arr
+                    + [
+                        f"{source_file}:: ERROR: CodeChat renderer - unable to find the HTML output file {base_html_file}."
+                    ],
+                )
+
+        # Recall that time is measured in seconds since the epoch,
+        # so that larger = newer.
+        try:
+            if html_file.stat().st_mtime > source_file.stat().st_mtime:
+                return str(html_file), []
+            else:
+                return (
+                    str(html_file),
+                    error_arr
+                    + [
+                        f"{source_file}:: ERROR: CodeChat renderer - source file newer than the HTML file {html_file}."
+                    ],
+                )
+        except OSError as e:
+            return (
+                str(html_file),
+                error_arr
+                + [
+                    f"{source_file}:: ERROR: CodeChat renderer - unable to check modification time of the HTML file {html_file}: {e}."
+                ],
+            )
+
+    def load_pretext_mapping(self):
+        return json.loads((self.output_path / "mapping.json").read_text())
+
+
 # Convert an project using an external renderer.
 async def _render_external_project(
     text: str, file_path_: str, _tool_or_project_path: str, co_build: Co_Build
 ) -> Tuple[str, str]:
     # Run from the directory containing the project file.
-    tool_or_project_path = Path(_tool_or_project_path)
-    project_path = tool_or_project_path.parent
+    project_conf_file_path = Path(_tool_or_project_path)
     await co_build(
-        "Loading project file {}.\n".format(tool_or_project_path),
+        f"Loading project file {project_conf_file_path}.\n",
     )
 
     # Read the project configuration.
     try:
-        with open(tool_or_project_path, encoding="utf-8") as f:
-            data = f.read()
-    except Exception as e:
-        return "", "{}:: ERROR: Unable to open. {}".format(tool_or_project_path, e)
+        project_conf = ProjectConfFile(project_conf_file_path)
+    except RuntimeError as e:
+        return "", str(e)
 
-    schema = strictyaml.Map(
-        {
-            strictyaml.Optional("source_path", default="."): strictyaml.Str(),
-            "output_path": strictyaml.Str(),
-            "args": strictyaml.Str() | strictyaml.Seq(strictyaml.Str()),
-            strictyaml.Optional("html_ext", default=".html"): strictyaml.Str(),
-        }
-    )
-    try:
-        data_dict = strictyaml.load(data, schema).data
-    except strictyaml.YAMLError as e:
-        return "", "{}:: ERROR: Unable to parse. {}".format(tool_or_project_path, e)
-
-    # Make paths absolute.
-    def abs_path(path: Union[str, Path]) -> Path:
-        path_ = Path(path)
-        if not path_.is_absolute():
-            path_ = project_path / path_
-        return path_
-
-    source_path = abs_path(data_dict["source_path"])
-    output_path = abs_path(data_dict["output_path"])
     file_path = Path(file_path_)
 
-    # Determine first guess at the location of the rendered HTML.
-    error_arr = []
-    try:
-        base_html_file = output_path / file_path.relative_to(source_path)
-    except ValueError as e:
-        # Give some arbitrary value to the output path, since it can't be determined.
-        base_html_file = output_path
-        error_arr.append(
-            "{}:: ERROR: unable to compute path relative to {}. {}".format(
-                file_path, source_path, e
-            )
-        )
-
     # Compare dates to see if the rendered file is current
-    html_ext = data_dict["html_ext"]
-    html_file, error = _checkModificationTime(file_path, base_html_file, html_ext)
+    html_file, error_arr = project_conf.checkModificationTime(file_path)
 
     # If not, render and try again.
-    if error:
-        # Perform replacement on the args.
-        def args_format(arg):
-            return arg.format(
-                project_path=project_path,
-                source_path=source_path,
-                output_path=output_path,
-            )
-
-        args = data_dict["args"]
-        args = (
-            args_format(args)
-            if isinstance(args, str)
-            else [args_format(arg) for arg in args]
-        )
+    if error_arr:
         # Render.
-        stdout, stderr = await _run_subprocess(args, project_path, None, True, co_build)
-        html_file, error = _checkModificationTime(file_path, base_html_file, html_ext)
-        if error:
-            error_arr.append(error)
+        stdout, stderr = await _run_subprocess(
+            project_conf.args, project_conf_file_path.parent, None, True, co_build
+        )
+        html_file, error_arr = project_conf.checkModificationTime(file_path)
     else:
         stderr = ""
 
@@ -340,46 +470,6 @@ def _dummy_context_manager() -> Generator:
     yield
 
 
-# _`_checkModificationTime`: Return False if source_file is newer than output_file; otherwise, return string with an error message.
-def _checkModificationTime(
-    source_file: Path, base_html_file: Path, html_ext: str
-) -> Tuple[str, str]:
-
-    # Look for the resulting HTML.
-    possible_html_file = base_html_file.with_suffix(html_ext)
-    if possible_html_file.exists():
-        html_file = possible_html_file
-    else:
-        possible_html_file = Path(str(base_html_file) + html_ext)
-        if possible_html_file.exists():
-            html_file = possible_html_file
-        else:
-            return (
-                str(base_html_file),
-                f"{source_file}:: ERROR: CodeChat renderer - unable to find the HTML output file {base_html_file}.",
-            )
-
-    # Recall that time is measured in seconds since the epoch,
-    # so that larger = newer.
-    try:
-        if html_file.stat().st_mtime > source_file.stat().st_mtime:
-            return str(html_file), ""
-        else:
-            return (
-                str(html_file),
-                "{}:: ERROR: CodeChat renderer - source file older than the HTML file {}.".format(
-                    source_file, html_file
-                ),
-            )
-    except OSError as e:
-        return (
-            str(html_file),
-            "{}:: ERROR: CodeChat renderer - unable to check modification time of the html file {}: {}.".format(
-                source_file, html_file, e
-            ),
-        )
-
-
 # Run a subprocess, optionally streaming the stdout.
 async def _run_subprocess(
     args: Union[List[str], str],
@@ -559,6 +649,8 @@ async def render_file(
 ) -> Tuple[
     # was_performed: True if the render was performed. False if this is a project and the source file is dirty; in this case, the render is skipped.
     bool,
+    # project_path: If this is a project, then ``project_path``` contains a path to the CodeChat project configuration file; otherwise, ``project_path`` is ``None``.
+    Optional[str],
     # rendered_file_path: A path to the rendered file.
     #
     # - If this is a project, the rendered file is different from ``file_path``, since it points to the location on disk where the external renderer wrote the HTML. In this case, the ``html`` return value is ``None``, since the HTMl should be read from the disk instead.
@@ -574,7 +666,7 @@ async def render_file(
 
     # Projects require a clean file in order to render.
     if is_project and is_dirty:
-        return False, "", None, ""
+        return False, None, "", None, ""
 
     if asyncio.iscoroutinefunction(renderer):
         # Coroutines get the queue, so they can report progress during the build.
@@ -588,7 +680,13 @@ async def render_file(
     # Update the client's state, now that the rendering is complete.
     if is_project:
         # For projects, the rendered HTML is already on disk; a path to this rendered file is returned.
-        return True, html_string_or_file_path, None, err_string
+        return (
+            True,
+            str(tool_or_project_path),
+            html_string_or_file_path,
+            None,
+            err_string,
+        )
     else:
         # Otherwise, the rendered HTML is returned as a string and can be directly used. Provide a path to the source file which was just rendered.
-        return True, file_path, html_string_or_file_path, err_string
+        return True, None, file_path, html_string_or_file_path, err_string
